/*******************************
Global Statistical Significance (GloSS) filter
Program written by Filippo Radicchi
*********************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_fft_real.h>
#include <gsl/gsl_fft_halfcomplex.h>
#include <gsl/gsl_spline.h>


void count_nodes(char *filename, int *MM);
void read_netw(char *filename, int **edges, int *ww);
int directed_stats(int **edges, int *ww, int E, int *kin, int *kout, int *sin, int *sout, int N);
int undirected_stats(int **edges, int *ww, int E, int *kin, int *kout, int *sin, int *sout, int N);
void calculate_distr_weights(int E, int *ww, double *xx, double *prob, int Q);
int determine_Q (int *sin, int *sout, int N);
void calculate_power(int N, int k, double *prob, double *prod, double **table);
void calculate_inverse_fourier(double **table, int N, int k, int *deg_tr);
double pvalue(int s1, int k1, int s2, int k2, int w, double **table, int N);
void calculate_pvalue(int **edges, int *ww, int E, int *kin, int *kout, int *sin, int *sout, int num_points, double **table, double *pp);
void print_table(char *filename, double **table, double *xx, int k, int N);
void print_rank(char *filedata, int **edges, int *ww, double *pp, int E);






void print_msg(char *text)
{
  printf("%s",text);
  fflush(stdout);
}




int main (int argc, char **argv)
{

  if(argc!=3) {
    printf("\n#########\n\n");
    printf("The correct usage is:\n./GloSS_discrete edge_list d\n\n");
    printf("edge_list should be a three columns file listing in start node, end node and weight of the connection\n");
    printf("d can be set equal to 1 for the analysis of undirected networks ");
    printf("or 2 for directed ones\n");
    printf("\n#########\n\n\n");
    exit(0);
  }

  int DIR=atoi(argv[2]);
  if(DIR!=1 && DIR!=2) {
    printf("\n#########\n\n");
    printf("The second argument must be 1 or 2, depending if you want analyze an undirected or directed graph\n");
    printf("\n#########\n\n\n");   
    exit(0);
  }



  printf("\n#########\n\n");
  printf("Global Statistical Significance (GloSS) filter\n");
  if(DIR==1) printf("\nReading the undirected network %s\n\n",argv[1]);
  if(DIR==2) printf("\nReading the directed network %s\n\n",argv[1]);



  char filename[100];
  int k;
  int MM[2];
  int N, E;
  count_nodes(argv[1], MM);
  N=MM[0];
  E=MM[1];
  printf("## Tot nodes %d\n",N);
  printf("## Tot edges %d\n",E);
  int **edges, i;
  int *ww;
  double *pp;
  ww=(int *)malloc((E+1)*sizeof(int));
  pp=(double *)malloc((E+1)*sizeof(double));
  edges=(int **)malloc((E+1)*sizeof(int *));
  for(i=1;i<=E;i++) edges[i]=(int *)malloc(2*sizeof(int));
  read_netw(argv[1], edges, ww);




  int *kin, *kout, *sin, *sout, kmax;
  kin=(int *)malloc((N+1)*sizeof(int));
  kout=(int *)malloc((N+1)*sizeof(int));
  sin=(int *)malloc((N+1)*sizeof(int));
  sout=(int *)malloc((N+1)*sizeof(int));
  

  if(DIR==1) kmax=undirected_stats(edges, ww, E, kin, kout, sin, sout, N);
  if(DIR==2) kmax=undirected_stats(edges, ww, E, kin, kout, sin, sout, N);
  printf("## kmax %d\n",kmax);






  int *deg_tr;
  deg_tr=(int *)malloc((int)kmax*sizeof(int));
  for(i=0;i<kmax;i++) deg_tr[i]=1;
  for(i=1;i<=N;i++) 
    {
      if(kin[i]>1) deg_tr[kin[i]-2]++;
      if(kout[i]>1) deg_tr[kout[i]-2]++;
      
    }

 
 
 

  
 

  int Q, j;

  Q=determine_Q (sin, sout, N);

  if(Q>23)
    {
      printf("\n\n!!!!\n");
      printf("If you want to be more accurate you should increment the resolution parameter Q\nNotice that high values of Q require high memory usage\n");
      printf("!!!!\n\n\n");
    }

  printf("## parameter Q %d\n",Q);
  
  int num_points=(int)pow(2.0,(double)Q);


  double *prob, *xx, dx=1.0;
  prob=(double *)malloc(num_points*sizeof(double));
  xx=(double *)malloc(num_points*sizeof(double));
  

  calculate_distr_weights(E, ww, xx, prob, Q);





  
  print_msg("## Calculating Fourier Transform : ");
  gsl_fft_real_radix2_transform(prob, 1, num_points);
  print_msg("done\n");


  double *prod, **table;
  prod=(double *)malloc(num_points*sizeof(double));
  table=(double **)malloc(kmax*sizeof(double *));
  for(i=0;i<kmax;i++) 
    {
      table[i]=(double *)malloc(num_points*sizeof(double)); 
      for(j=0;j<num_points;j++) table[i][j]=0.0;
    } 







  print_msg("## Calculating Powers : ");
  calculate_power(num_points, kmax, prob, prod, table);
  print_msg("done\n");



  

  print_msg("## Calculating Inverse Fourier Transform : ");
  calculate_inverse_fourier(table, num_points, kmax, deg_tr);
  print_msg("done\n");


  print_msg("## Calculating p-values : ");
  calculate_pvalue(edges, ww, E, kin, kout, sin, sout, num_points, table, pp);
  print_msg("done\n");

  print_msg("## Writing file : ");
  print_rank(argv[1], edges, ww, pp, E);
  print_msg("done\n\n");



  for(i=1;i<=E;i++) free(edges[i]);
  free(edges);
  free(ww);
  free(kin);
  free(kout);
  free(sin);
  free(sout);
  free(prob);
  free(xx);
  for(i=0;i<kmax;i++) free(table[i]);
  free(table);
  free(prod);
  free(pp);
  free(deg_tr);

  return 0;
}















double pvalue(int s1, int k1, int s2, int k2, int w, double **table, int N)
{
 
  int i;
 
 
  int q1, q2;
 
  double cc=0.0, bb=0.0;
  for(i=0;i<N;i++)
    {
      q1=s1-i;
      q2=s2-i;
     

      //printf("w %d   s1 %d k1 %d    s2 %d k2 %d    i %d   q1 %d   q2 %d: %g %g %g\n",w,s1,k1,s2,k2,i,q1,q2,table[0][i],table[k1-2][q1],table[k2-2][q2]);

      if(q1>=1 && q2>=1)
	{	  
	  if(i>=w) cc+=table[0][i]*table[k1-2][q1]*table[k2-2][q2]; 
	  bb+=table[0][i]*table[k1-2][q1]*table[k2-2][q2]; 
	}

      
      else{goto endloop;}
    }
  endloop:
  

  //printf("%g %g  %g\n\n",cc,bb,cc/bb);
  
  

  if(bb>0.0) return cc/bb;
  return -10.0;
}





void calculate_pvalue(int **edges, int *ww, int E, int *kin, int *kout, int *sin, int *sout, int num_points, double **table, double *pp)
{
 
  int s1, s2, w;
  int i, k1, k2, n1, n2;
  int C=0;
  for(i=1;i<=E;i++)
    {
      n1=edges[i][0];
      n2=edges[i][1];
      k1=kout[n1];
      k2=kin[n2];
      s1=sout[n1];
      s2=sin[n2];
      w=ww[i];
      pp[i]=1.0;
      if(k1>1 && k2>1) pp[i]=pvalue(s1, k1, s2, k2, w, table, num_points);
      if(pp[i]<=0.01) C++;
    }


  //printf("C %d   %g\n",C,(double)C/(double)E);
}









void calculate_power(int N, int k, double *prob, double *prod, double **table)
{
  int i, r, v;
  double re1, im1, re2, im2, a, b, c, d;
  for(i=0;i<N;i++) 
    {
      prod[i]=prob[i];
      //printf("0 %d %g\n",i,prod[i]);
    }
  
  for(i=0;i<N;i++) table[0][i]=prod[i];
   
  for(v=1;v<k;v++)
    {

      for(i=0;i<=N/2;i++)
	{
	  r=N-i;

	  //printf("%d %d\n",i,r);

	  re1=prob[i];
	  im1=0.0;
	  if(r<N && i!=N/2) im1=prob[r];
	 
	  re2=prod[i];
          im2=0.0;
          if(r<N && i!=N/2) im2=prod[r];


	  a=re1*re2;
	  b=im1*im2;
	  c=re1*im2;
	  d=re2*im1;


	  prod[i]=a-b;
	  if(r<N && r!=N/2) prod[r]=c+d;

	}
     
      for(i=0;i<N;i++) 
	{
	  table[v][i]=prod[i];
	  //printf("%d %d %g\n",v,i,table[v][i]);
	}
      
    }
 
}




void calculate_inverse_fourier(double **table, int N, int k,  int *deg_tr)
{
  int i, j;
  double norm;
  for(i=0;i<k;i++) 
    {
      if(deg_tr[i]>0){
	gsl_fft_halfcomplex_radix2_inverse (table[i], 1, N);
	norm=0.0;
	for(j=0;j<N;j++) if(table[i][j]<0.0) table[i][j]=0.0;
	for(j=0;j<N;j++) norm+=table[i][j];
	for(j=0;j<N;j++) table[i][j]=table[i][j]/norm;
      }
    }
}
			       






void count_nodes(char *filename, int *MM)
{
  int q, i, j, N=0, E=0;
  double k;
  FILE *f;
  f=fopen(filename,"r");
  while(!feof(f))
    {
      q=fscanf(f,"%d %d %lf",&i,&j,&k);
      if(q<=0) goto exitfile;
      if(i>N) N=i;
      if(j>N) N=j;
      E++;
    }
 exitfile:
  fclose(f);
  MM[0]=N;
  MM[1]=E;
}



///////////////////////////////////////////////////////

void read_netw(char *filename, int **edges, int *ww)
{
  int q, i, j, k, E=0;
 
  FILE *f;
  f=fopen(filename,"r");
  while(!feof(f))
    {
      q=fscanf(f,"%d %d %d",&i,&j,&k);
      if(q<=0) goto exitfile;
     

      E++;
      edges[E][0]=i;
      edges[E][1]=j;
      ww[E]=k;

    

    }
 exitfile:
  fclose(f);
 
}




int undirected_stats(int **edges, int *ww, int E, int *kin, int *kout, int *sin, int *sout, int N)
{
  int kmax=0;
  int i, n1, n2;
  for(i=1;i<=N;i++) kin[i]=kout[i]=sin[i]=sout[i]=0;


  for(i=1;i<=E;i++)
    {
      n1=edges[i][0];
      n2=edges[i][1];

      kin[n1]++;
      kout[n1]++;

      kin[n2]++;
      kout[n2]++;

      sin[n1]+=ww[i];
      sout[n1]+=ww[i];

      sin[n2]+=ww[i];
      sout[n2]+=ww[i];
    }
  for(i=1;i<=N;i++) if(kin[i]>kmax) kmax=kin[i];
  for(i=1;i<=N;i++) if(kout[i]>kmax) kmax=kout[i];
  return kmax;
}


int directed_stats(int **edges, int *ww, int E, int *kin, int *kout, int *sin, int *sout, int N)
{
  int kmax=0.0;
  int i, n1, n2;
  for(i=1;i<=N;i++) kin[i]=kout[i]=sin[i]=sout[i]=0;
  for(i=1;i<=E;i++)
    {
      n1=edges[i][0];
      n2=edges[i][1];

     
      kout[n1]++;

      kin[n2]++;



      sout[n1]+=ww[i];

      sin[n2]+=ww[i];

    }
  for(i=1;i<=N;i++) if(kin[i]>kmax) kmax=kin[i];
  for(i=1;i<=N;i++) if(kout[i]>kmax) kmax=kout[i];
  return kmax;
}



int determine_Q (int *sin, int *sout, int N)
{
  int i;
  int smax=0;
  for(i=1;i<=N;i++)
    {
      if(sin[i]>smax) smax=sin[i];
      if(sout[i]>smax) smax=sout[i];
    }
  int Q=1;
  while(Q)
    {
      if(pow(2.0,(double)Q) > smax) goto next;
      Q++;
    }
 next:
  printf("smax %d\n",smax);
  return Q;
}



void calculate_distr_weights(int E, int *ww, double *xx, double *prob, int Q)
{
  
  int i, num_points=(int)(pow(2.0,(double)Q));



  xx[0]=0.0;
  for(i=1;i<num_points;i++) xx[i]=xx[i-1]+1.0;
  for(i=0;i<num_points;i++) prob[i]=0.0;

  for(i=1;i<=E;i++) prob[ww[i]]+=1.0;
  for(i=0;i<num_points;i++) prob[i]/=(double)E;

 
}







void print_table(char *filename, double **table, double *xx, int k, int N)
{
  FILE *f;
  f=fopen(filename,"w");
  int i;
  for(i=0;i<N;i++) 
    {
      //printf("%d %g %g\n",k,xx[i],table[k-1][i]);
      //fflush(stdout);
      fprintf(f,"%g ",xx[i]);
      fprintf(f,"%g\n",table[k-1][i]);
    }

  fclose(f);
}



void print_rank(char *filedata, int **edges, int *ww, double *pp, int E)
{
  FILE *f;
  int i;
  char filename[100];

  /*sprintf(filename,"DATA/rank_%s",filedata);
  f=fopen(filename,"w");
  for(i=1;i<=E;i++) fprintf(f,"%d %d %g\n",edges[i][0],edges[i][1],pp[i]);
  fclose(f);
  */

  sprintf(filename,"GloSS_%s",filedata);
  f=fopen(filename,"w");
  for(i=1;i<=E;i++) fprintf(f,"%d %d %g %d\n",edges[i][0],edges[i][1],pp[i],ww[i]);
  fclose(f);

} 




